from __future__ import annotations

from collections import defaultdict
from dataclasses import field, dataclass
from typing import Self

from ..datastructures.prefixed_name import PrefixedName
from ..robots.abstract_robot import (
    Finger,
    ParallelGripper,
    Arm,
    Camera,
    FieldOfView,
    Neck,
    AbstractRobot,
)
from .robot_mixins import HasNeck, SpecifiesLeftRightArm
from ..spatial_types import Quaternion, Vector3
from ..world import World


@dataclass
class Tracy(AbstractRobot, SpecifiesLeftRightArm, HasNeck):
    """
    Represents two UR10e Arms on a table, with a pole between them holding a small camera.
     Example can be found at: https://vib.ai.uni-bremen.de/page/comingsoon/the-tracebot-laboratory/
    """

    def __hash__(self):
        return hash(
            tuple(
                [self.__class__]
                + sorted([kse.name for kse in self.kinematic_structure_entities])
            )
        )

    def load_srdf(self): ...

    @classmethod
    def from_world(cls, world: World) -> Self:
        """
        Creates a Tracy robot semantic annotation from the given world.

        :param world: The world from which to create the robot semantic annotation.

        :return: A Tracy robot semantic annotation.
        """
        with world.modify_world():
            robot = cls(
                name=PrefixedName(name="tracy", prefix=world.name),
                root=world.get_body_by_name("table"),
                _world=world,
            )

            # Create left arm
            left_gripper_thumb = Finger(
                name=PrefixedName("left_gripper_thumb", prefix=robot.name.name),
                root=world.get_body_by_name("left_robotiq_85_left_knuckle_link"),
                tip=world.get_body_by_name("left_robotiq_85_left_finger_tip_link"),
                _world=world,
            )

            left_gripper_finger = Finger(
                name=PrefixedName("left_gripper_finger", prefix=robot.name.name),
                root=world.get_body_by_name("left_robotiq_85_right_knuckle_link"),
                tip=world.get_body_by_name("left_robotiq_85_right_finger_tip_link"),
                _world=world,
            )

            left_gripper = ParallelGripper(
                name=PrefixedName("left_gripper", prefix=robot.name.name),
                root=world.get_body_by_name("left_robotiq_85_base_link"),
                tool_frame=world.get_body_by_name("l_gripper_tool_frame"),
                front_facing_orientation=Quaternion(0.5, 0.5, 0.5, 0.5),
                front_facing_axis=Vector3(0, 0, 1),
                thumb=left_gripper_thumb,
                finger=left_gripper_finger,
                _world=world,
            )
            left_arm = Arm(
                name=PrefixedName("left_arm", prefix=robot.name.name),
                root=world.get_body_by_name("table"),
                tip=world.get_body_by_name("left_wrist_3_link"),
                manipulator=left_gripper,
                _world=world,
            )

            robot.add_arm(left_arm)

            right_gripper_thumb = Finger(
                name=PrefixedName("right_gripper_thumb", prefix=robot.name.name),
                root=world.get_body_by_name("right_robotiq_85_left_knuckle_link"),
                tip=world.get_body_by_name("right_robotiq_85_left_finger_tip_link"),
                _world=world,
            )
            right_gripper_finger = Finger(
                name=PrefixedName("right_gripper_finger", prefix=robot.name.name),
                root=world.get_body_by_name("right_robotiq_85_right_knuckle_link"),
                tip=world.get_body_by_name("right_robotiq_85_right_finger_tip_link"),
                _world=world,
            )
            right_gripper = ParallelGripper(
                name=PrefixedName("right_gripper", prefix=robot.name.name),
                root=world.get_body_by_name("right_robotiq_85_base_link"),
                tool_frame=world.get_body_by_name("r_gripper_tool_frame"),
                front_facing_orientation=Quaternion(0.5, 0.5, 0.5, 0.5),
                front_facing_axis=Vector3(0, 0, 1),
                thumb=right_gripper_thumb,
                finger=right_gripper_finger,
                _world=world,
            )
            right_arm = Arm(
                name=PrefixedName("right_arm", prefix=robot.name.name),
                root=world.get_body_by_name("table"),
                tip=world.get_body_by_name("right_wrist_3_link"),
                manipulator=right_gripper,
                _world=world,
            )
            robot.add_arm(right_arm)

            camera = Camera(
                name=PrefixedName("camera", prefix=robot.name.name),
                root=world.get_body_by_name("camera_link"),
                forward_facing_axis=Vector3(0, 0, 1),
                field_of_view=FieldOfView(horizontal_angle=1.047, vertical_angle=0.785),
                minimal_height=0.8,
                maximal_height=1.7,
                _world=world,
            )

            # Probably should be classified as "Neck", as that implies that i can move.
            neck = Neck(
                name=PrefixedName("neck", prefix=robot.name.name),
                sensors={camera},
                root=world.get_body_by_name("camera_pole"),
                tip=world.get_body_by_name("camera_link"),
                _world=world,
            )

            robot.add_kinematic_chain(neck)
            world.add_semantic_annotation(robot, exists_ok=True)

            vel_limits = defaultdict(lambda: 0.2)
            robot.tighten_dof_velocity_limits_of_1dof_connections(new_limits=vel_limits)

            return robot



from dataclasses import dataclass, field
from os.path import dirname

from typing_extensions import Optional, List, Dict, Any, Type, Callable, ClassVar

from ..world import World
from ..world_description.world_entity import SemanticAnnotation
from .reasoner import CaseReasoner


@dataclass
class WorldReasoner:
    """
    A utility class that uses CaseReasoner for reasoning on the world concepts.
    """

    world: World
    """
    The world instance to reason on.
    """
    _last_world_model_version: Optional[int] = field(init=False, default=None)
    """
    The last world model version of the world used when :py:meth:`reason` 
    was last called.
    """
    reasoner: CaseReasoner = field(init=False)
    """
    The case reasoner that is used to reason on the world concepts.
    """
    model_directory: ClassVar[str] = dirname(__file__)
    """
    The directory where the rdr model folder is located.
    """

    def __post_init__(self):
        self.reasoner = CaseReasoner(self.world, model_directory=self.model_directory)

    def infer_semantic_annotations(self) -> List[SemanticAnnotation]:
        """
        Infer the semantic annotations of the world by calling the :py:meth:`reason` method and extracting all inferred semantic annotations.

        :return: The inferred semantic annotations of the world.
        """
        with self.world.modify_world():
            result = self.reason()
        return result.get("semantic_annotations", [])

    def reason(self) -> Dict[str, Any]:
        """
        Perform rule-based reasoning on the current world and infer all possible concepts.

        :return: The inferred concepts as a dictionary mapping concept name to all inferred values of that concept.
        """
        if self.world._model_version != self._last_world_model_version:
            self.reasoner.result = self.reasoner.rdr.classify(self.world)
            self._update_world_attributes()
            self._last_world_model_version = self.world._model_version
        return self.reasoner.result

    def _update_world_attributes(self):
        """
        Update the world attributes from the values in the result of the latest :py:meth:`reason` call.
        """
        for attr_name, attr_value in self.reasoner.result.items():
            if isinstance(getattr(self.world, attr_name), list):
                attr_value = list(attr_value)
            if attr_name == "semantic_annotations":
                for semantic_annotation in attr_value:
                    self.world.add_semantic_annotation(
                        semantic_annotation, exists_ok=True
                    )
            else:
                setattr(self.world, attr_name, attr_value)

    def fit_semantic_annotations(
        self,
        required_semantic_annotations: List[Type[SemanticAnnotation]],
        update_existing_semantic_annotations: bool = False,
        world_factory: Optional[Callable] = None,
        scenario: Optional[Callable] = None,
    ) -> None:
        """
        Fit the world RDR to the required semantic annotation types.

        :param required_semantic_annotations: A list of semantic annotation types that the RDR should be fitted to.
        :param update_existing_semantic_annotations: If True, existing semantic annotations will be updated with new rules, else they will be skipped.
        :param world_factory: Optional callable that can be used to recreate the world object.
        :param scenario: Optional callable that represents the test method or scenario that is being executed.
        """
        self.reasoner.fit_attribute(
            "semantic_annotations",
            required_semantic_annotations,
            False,
            update_existing_rules=update_existing_semantic_annotations,
            case_factory=world_factory,
            scenario=scenario,
        )



from __future__ import annotations

import itertools
import tempfile
from abc import ABC, abstractmethod
from dataclasses import dataclass, field, fields
from functools import cached_property

import numpy as np
import trimesh
import trimesh.exchange.stl
from random_events.interval import SimpleInterval, Bound, closed
from random_events.product_algebra import SimpleEvent
from random_events.utils import SubclassJSONSerializer
from typing_extensions import Optional, List, TYPE_CHECKING, Dict, Any, Self, Tuple

from ..datastructures.variables import SpatialVariables
from ..spatial_types import TransformationMatrix, Point3
from ..spatial_types.spatial_types import Expression
from ..spatial_types.symbol_manager import symbol_manager
from ..utils import IDGenerator

id_generator = IDGenerator()


def transformation_from_json(data: Dict[str, Any]) -> TransformationMatrix:
    """
    Creates a transformation matrix from a JSON-compatible dictionary.

    Use this together with `transformation_to_json`.

    This is needed since SpatialTypes cannot inherit from SubClassJSONSerializer.
    They can't inherit since the conversion to JSON needs the symbol_manager, which would cause a cyclic dependency.
    """
    return TransformationMatrix.from_xyz_quaternion(
        *data["position"][:3], *data["quaternion"]
    )


def transformation_to_json(transformation: TransformationMatrix) -> Dict[str, Any]:
    """
    Converts a transformation matrix to a JSON-compatible dictionary.

    Use this together with `transformation_from_json`.

    This is needed since SpatialTypes cannot inherit from SubClassJSONSerializer.
    They can't inherit since the conversion to JSON needs the symbol_manager, which would cause a cyclic dependency.
    """
    position = symbol_manager.evaluate_expr(transformation.to_position()).tolist()
    quaternion = symbol_manager.evaluate_expr(transformation.to_quaternion()).tolist()
    return {"position": position, "quaternion": quaternion}


@dataclass
class Color(SubclassJSONSerializer):
    """
    Dataclass for storing rgba_color as an RGBA value.
    The values are stored as floats between 0 and 1.
    The default rgba_color is white.
    """

    R: float = 1.0
    """
    Red value of the color.
    """

    G: float = 1.0
    """
    Green value of the color.
    """

    B: float = 1.0
    """
    Blue value of the color.
    """

    A: float = 1.0
    """
    Opacity of the color.
    """

    def __post_init__(self):
        """
        Make sure the color values are floats, because ros2 sucks.
        """
        self.R = float(self.R)
        self.G = float(self.G)
        self.B = float(self.B)
        self.A = float(self.A)

    def to_json(self) -> Dict[str, Any]:
        return {**super().to_json(), "R": self.R, "G": self.G, "B": self.B, "A": self.A}

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        return cls(R=data["R"], G=data["G"], B=data["B"], A=data["A"])

    def to_rgba(self) -> Tuple[float, float, float, float]:
        return (self.R, self.G, self.B, self.A)


@dataclass
class Scale(SubclassJSONSerializer):
    """
    Dataclass for storing the scale of geometric objects.
    """

    x: float = 1.0
    """
    The scale in the x direction.
    """

    y: float = 1.0
    """
    The scale in the y direction.
    """

    z: float = 1.0
    """
    The scale in the z direction.
    """

    def to_json(self) -> Dict[str, Any]:
        return {**super().to_json(), "x": self.x, "y": self.y, "z": self.z}

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        return cls(x=data["x"], y=data["y"], z=data["z"])

    def __post_init__(self):
        """
        Make sure the scale values are floats, because ros2 sucks.
        """
        self.x = float(self.x)
        self.y = float(self.y)
        self.z = float(self.z)

    @property
    def simple_event(self) -> SimpleEvent:
        return SimpleEvent(
            {
                SpatialVariables.x.value: closed(-self.x / 2, self.x / 2),
                SpatialVariables.y.value: closed(-self.y / 2, self.y / 2),
                SpatialVariables.z.value: closed(-self.z / 2, self.z / 2),
            }
        )


@dataclass
class Shape(ABC, SubclassJSONSerializer):
    """
    Base class for all shapes in the world.
    """

    origin: TransformationMatrix = field(default_factory=TransformationMatrix)

    color: Color = field(default_factory=Color)

    @property
    @abstractmethod
    def local_frame_bounding_box(self) -> BoundingBox:
        """
        Returns the bounding box of the shape
        """

    @property
    @abstractmethod
    def mesh(self) -> trimesh.Trimesh:
        """
        The mesh object of the shape.
        This should be implemented by subclasses.
        """

    def to_json(self) -> Dict[str, Any]:
        return {
            **super().to_json(),
            "origin": transformation_to_json(self.origin),
            "color": self.color.to_json(),
        }

    def __eq__(self, other: Shape) -> bool:
        """Custom equality comparison that handles TransformationMatrix equivalence"""
        if not isinstance(other, self.__class__):
            return False

        # Get all field names from the dataclass
        field_names = [f.name for f in fields(self)]

        for field_name in field_names:
            self_value = getattr(self, field_name)
            other_value = getattr(other, field_name)

            if field_name != "origin":
                if self_value != other_value:
                    return False
        if not np.allclose(self.origin.to_np(), other.origin.to_np()):
            return False

        return True


@dataclass(eq=False)
class Mesh(Shape, ABC):
    """
    Abstract mesh class.
    Subclasses must provide a `mesh` property returning a trimesh.Trimesh.
    """

    scale: Scale = field(default_factory=Scale)
    """
    Scale of the mesh.
    """

    @property
    @abstractmethod
    def mesh(self) -> trimesh.Trimesh:
        """Return the loaded mesh object."""
        raise NotImplementedError

    @property
    def local_frame_bounding_box(self) -> BoundingBox:
        """
        Returns the local bounding box of the mesh.
        The bounding box is axis-aligned and centered at the origin.
        """
        return BoundingBox.from_mesh(self.mesh, self.origin)

    def to_json(self) -> Dict[str, Any]:
        return {
            **super().to_json(),
            "mesh": self.mesh.to_dict(),
            "scale": self.scale.to_json(),
        }

    @classmethod
    @abstractmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self: ...


@dataclass(eq=False)
class FileMesh(Mesh):
    """
    A mesh shape defined by a file.
    """

    filename: str = ""
    """
    Filename of the mesh.
    """

    @cached_property
    def mesh(self) -> trimesh.Trimesh:
        """
        The mesh object.
        """
        mesh = trimesh.load_mesh(self.filename)
        mesh.visual.vertex_colors = trimesh.visual.color.to_rgba(self.color.to_rgba())
        return mesh

    def to_json(self) -> Dict[str, Any]:
        json = {
            **super().to_json(),
            "mesh": self.mesh.to_dict(),
            "scale": self.scale.to_json(),
        }
        json["type"] = json["type"].replace("FileMesh", "TriangleMesh")
        return json

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        raise NotImplementedError(
            f"{cls} does not support loading from JSON due to filenames across different systems."
            f" Use TriangleMesh instead."
        )


@dataclass(eq=False)
class TriangleMesh(Mesh):
    """
    A mesh shape defined by vertices and faces.
    """

    mesh: Optional[trimesh.Trimesh] = None
    """
    The loaded mesh object.
    """

    @cached_property
    def file(self):
        f = tempfile.NamedTemporaryFile(delete=False)
        with open(f.name, "w") as fd:
            fd.write(trimesh.exchange.stl.export_stl_ascii(self.mesh))
        return f

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> TriangleMesh:
        mesh = trimesh.Trimesh(
            vertices=data["mesh"]["vertices"], faces=data["mesh"]["faces"]
        )
        origin = transformation_from_json(data["origin"])
        scale = Scale.from_json(data["scale"])
        return cls(mesh=mesh, origin=origin, scale=scale)


@dataclass(eq=False)
class Sphere(Shape):
    """
    A sphere shape.
    """

    radius: float = 0.5
    """
    Radius of the sphere.
    """

    @property
    def mesh(self) -> trimesh.Trimesh:
        """
        Returns a trimesh object representing the sphere.
        """
        mesh = trimesh.creation.icosphere(subdivisions=2, radius=self.radius)
        mesh.visual.vertex_colors = trimesh.visual.color.to_rgba(self.color.to_rgba())
        return mesh

    @property
    def local_frame_bounding_box(self) -> BoundingBox:
        """
        Returns the bounding box of the sphere.
        """
        return BoundingBox(
            -self.radius,
            -self.radius,
            -self.radius,
            self.radius,
            self.radius,
            self.radius,
            self.origin,
        )

    def to_json(self) -> Dict[str, Any]:
        return {**super().to_json(), "radius": self.radius}

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        return cls(
            radius=data["radius"],
            origin=transformation_from_json(data["origin"]),
            color=Color.from_json(data["color"]),
        )


@dataclass(eq=False)
class Cylinder(Shape):
    """
    A cylinder shape.
    """

    width: float = 0.5
    height: float = 0.5

    @property
    def mesh(self) -> trimesh.Trimesh:
        """
        Returns a trimesh object representing the cylinder.
        """
        mesh = trimesh.creation.cylinder(
            radius=self.width / 2, height=self.height, sections=16
        )
        mesh.visual.vertex_colors = trimesh.visual.color.to_rgba(self.color.to_rgba())
        return mesh

    @property
    def local_frame_bounding_box(self) -> BoundingBox:
        """
        Returns the bounding box of the cylinder.
        The bounding box is axis-aligned and centered at the origin.
        """
        half_width = self.width / 2
        half_height = self.height / 2
        return BoundingBox(
            -half_width,
            -half_width,
            -half_height,
            half_width,
            half_width,
            half_height,
            self.origin,
        )

    def to_json(self) -> Dict[str, Any]:
        return {**super().to_json(), "width": self.width, "height": self.height}

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        return cls(
            width=data["width"],
            height=data["height"],
            origin=transformation_from_json(data["origin"]),
            color=Color.from_json(data["color"]),
        )


@dataclass(eq=False)
class Box(Shape):
    """
    A box shape. Pivot point is at the center of the box.
    """

    scale: Scale = field(default_factory=Scale)

    @property
    def mesh(self) -> trimesh.Trimesh:
        """
        Returns a trimesh object representing the box.
        The box is centered at the origin and has the specified scale.
        """
        mesh = trimesh.creation.box(extents=(self.scale.x, self.scale.y, self.scale.z))
        mesh.visual.vertex_colors = trimesh.visual.color.to_rgba(self.color.to_rgba())
        return mesh


    @property
    def local_frame_bounding_box(self) -> BoundingBox:
        """
        Returns the local bounding box of the box.
        The bounding box is axis-aligned and centered at the origin.
        """
        half_x = self.scale.x / 2
        half_y = self.scale.y / 2
        half_z = self.scale.z / 2
        return BoundingBox(
            -half_x,
            -half_y,
            -half_z,
            half_x,
            half_y,
            half_z,
            self.origin,
        )

    def to_json(self) -> Dict[str, Any]:
        return {**super().to_json(), "scale": self.scale.to_json()}

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        return cls(
            scale=Scale.from_json(data["scale"]),
            origin=transformation_from_json(data["origin"]),
            color=Color.from_json(data["color"]),
        )


@dataclass(eq=False)
class BoundingBox:
    min_x: float
    """
    The minimum x-coordinate of the bounding box.
    """

    min_y: float
    """
    The minimum y-coordinate of the bounding box.
    """

    min_z: float
    """
    The minimum z-coordinate of the bounding box.
    """

    max_x: float
    """
    The maximum x-coordinate of the bounding box.
    """

    max_y: float
    """
    The maximum y-coordinate of the bounding box.
    """

    max_z: float
    """
    The maximum z-coordinate of the bounding box.
    """

    origin: TransformationMatrix
    """
    The origin of the bounding box.
    """

    def __hash__(self):
        # The hash should be this since comparing those via hash is checking if those are the same and not just equal
        return hash(
            (self.min_x, self.min_y, self.min_z, self.max_x, self.max_y, self.max_z)
        )

    @property
    def x_interval(self) -> SimpleInterval:
        """
        :return: The x interval of the bounding box.
        """
        return SimpleInterval(self.min_x, self.max_x, Bound.CLOSED, Bound.CLOSED)

    @property
    def y_interval(self) -> SimpleInterval:
        """
        :return: The y interval of the bounding box.
        """
        return SimpleInterval(self.min_y, self.max_y, Bound.CLOSED, Bound.CLOSED)

    @property
    def z_interval(self) -> SimpleInterval:
        """
        :return: The z interval of the bounding box.
        """
        return SimpleInterval(self.min_z, self.max_z, Bound.CLOSED, Bound.CLOSED)

    @property
    def scale(self) -> Scale:
        """
        :return: The scale of the bounding box.
        """
        return Scale(self.depth, self.width, self.height)

    @property
    def depth(self) -> float:
        return self.max_x - self.min_x

    @property
    def height(self) -> float:
        return self.max_z - self.min_z

    @property
    def width(self) -> float:
        return self.max_y - self.min_y

    @property
    def simple_event(self) -> SimpleEvent:
        """
        :return: The bounding box as a random event.
        """
        return SimpleEvent(
            {
                SpatialVariables.x.value: self.x_interval,
                SpatialVariables.y.value: self.y_interval,
                SpatialVariables.z.value: self.z_interval,
            }
        )

    @property
    def dimensions(self) -> List[float]:
        """
        :return: The dimensions of the bounding box as a list [width, height, depth].
        """
        return [self.width, self.height, self.depth]

    def bloat(
        self, x_amount: float = 0.0, y_amount: float = 0, z_amount: float = 0
    ) -> BoundingBox:
        """
        Enlarges the bounding box by a given amount in all dimensions.

        :param x_amount: The amount to adjust minimum and maximum x-coordinates
        :param y_amount: The amount to adjust minimum and maximum y-coordinates
        :param z_amount: The amount to adjust minimum and maximum z-coordinates
        :return: New enlarged bounding box
        """
        return self.__class__(
            self.min_x - x_amount,
            self.min_y - y_amount,
            self.min_z - z_amount,
            self.max_x + x_amount,
            self.max_y + y_amount,
            self.max_z + z_amount,
            self.origin,
        )

    def contains(self, point: Point3) -> bool:
        """
        Check if the bounding box contains a point.
        """
        x, y, z = (
            (point.x.to_np(), point.y.to_np(), point.z.to_np())
            if isinstance(point.z, Expression)
            else (point.x, point.y, point.z)
        )

        return self.simple_event.contains((x, y, z))

    @classmethod
    def from_simple_event(cls, simple_event: SimpleEvent):
        """
        Create a list of bounding boxes from a simple random event.

        :param simple_event: The random event.
        :return: The list of bounding boxes.
        """
        result = []
        for x, y, z in itertools.product(
            simple_event[SpatialVariables.x.value].simple_sets,
            simple_event[SpatialVariables.y.value].simple_sets,
            simple_event[SpatialVariables.z.value].simple_sets,
        ):
            result.append(cls(x.lower, y.lower, z.lower, x.upper, y.upper, z.upper))
        return result

    def intersection_with(self, other: BoundingBox) -> Optional[BoundingBox]:
        """
        Compute the intersection of two bounding boxes.

        :param other: The other bounding box.
        :return: The intersection of the two bounding boxes or None if they do not intersect.
        """
        result = self.simple_event.intersection_with(other.simple_event)
        if result.is_empty():
            return None
        return self.__class__.from_simple_event(result)[0]

    def enlarge(
        self,
        min_x: float = 0.0,
        min_y: float = 0,
        min_z: float = 0,
        max_x: float = 0.0,
        max_y: float = 0.0,
        max_z: float = 0.0,
    ):
        """
        Enlarge the axis-aligned bounding box by a given amount in-place.
        :param min_x: The amount to enlarge the minimum x-coordinate
        :param min_y: The amount to enlarge the minimum y-coordinate
        :param min_z: The amount to enlarge the minimum z-coordinate
        :param max_x: The amount to enlarge the maximum x-coordinate
        :param max_y: The amount to enlarge the maximum y-coordinate
        :param max_z: The amount to enlarge the maximum z-coordinate
        """
        self.min_x -= min_x
        self.min_y -= min_y
        self.min_z -= min_z
        self.max_x += max_x
        self.max_y += max_y
        self.max_z += max_z

    def enlarge_all(self, amount: float):
        """
        Enlarge the axis-aligned bounding box in all dimensions by a given amount in-place.

        :param amount: The amount to enlarge the bounding box
        """
        self.enlarge(amount, amount, amount, amount, amount, amount)

    @classmethod
    def from_mesh(cls, mesh: trimesh.Trimesh, origin: TransformationMatrix) -> Self:
        """
        Create a bounding box from a trimesh object.
        :param mesh: The trimesh object.
        :param origin: The origin of the bounding box.
        :return: The bounding box.
        """
        bounds = mesh.bounds
        return cls(
            bounds[0][0],
            bounds[0][1],
            bounds[0][2],
            bounds[1][0],
            bounds[1][1],
            bounds[1][2],
            origin=origin,
        )

    def get_points(self) -> List[Point3]:
        """
        Get the 8 corners of the bounding box as Point3 objects.

        :return: A list of Point3 objects representing the corners of the bounding box.
        """
        return [
            Point3(x, y, z)
            for x in (self.min_x, self.max_x)
            for y in (self.min_y, self.max_y)
            for z in (self.min_z, self.max_z)
        ]

    @classmethod
    def from_min_max(cls, min_point: Point3, max_point: Point3) -> Self:
        """
        Set the axis-aligned bounding box from a minimum and maximum point.

        :param min_point: The minimum point
        :param max_point: The maximum point
        """
        assert min_point.reference_frame is not None
        assert (
            min_point.reference_frame == max_point.reference_frame
        ), "The reference frames of the minimum and maximum points must be the same."
        return cls(
            *min_point.to_np()[:3],
            *max_point.to_np()[:3],
            origin=TransformationMatrix(reference_frame=min_point.reference_frame),
        )

    def as_shape(self) -> Box:
        scale = Scale(
            x=self.max_x - self.min_x,
            y=self.max_y - self.min_y,
            z=self.max_z - self.min_z,
        )
        x = (self.max_x + self.min_x) / 2
        y = (self.max_y + self.min_y) / 2
        z = (self.max_z + self.min_z) / 2
        origin = TransformationMatrix.from_xyz_rpy(
            x, y, z, 0, 0, 0, self.origin.reference_frame
        )
        return Box(origin=origin, scale=scale)

    def transform_to_origin(self, reference_T_new_origin: TransformationMatrix) -> Self:
        """
        Transform the bounding box to a different reference frame.
        """
        origin_T_self = self.origin
        origin_frame = origin_T_self.reference_frame
        world = origin_frame._world

        reference_T_origin = world.compute_forward_kinematics(
            reference_T_new_origin.reference_frame, origin_frame
        )

        reference_T_self: TransformationMatrix = reference_T_origin @ origin_T_self

        # Get all 8 corners of the BB in link-local space
        list_self_T_corner = [
            TransformationMatrix.from_point_rotation_matrix(self_T_corner)
            for self_T_corner in self.get_points()
        ]  # shape (8, 3)

        list_reference_T_corner = [
            reference_T_self @ self_T_corner for self_T_corner in list_self_T_corner
        ]

        list_reference_P_corner = [
            reference_T_corner.to_position().to_np()[:3]
            for reference_T_corner in list_reference_T_corner
        ]

        # Compute world-space bounding box from transformed corners
        min_corner = np.min(list_reference_P_corner, axis=0)
        max_corner = np.max(list_reference_P_corner, axis=0)

        world_bb = BoundingBox.from_min_max(
            Point3.from_iterable(
                min_corner, reference_frame=reference_T_new_origin.reference_frame
            ),
            Point3.from_iterable(
                max_corner, reference_frame=reference_T_new_origin.reference_frame
            ),
        )

        return world_bb

    def __eq__(self, other: BoundingBox) -> bool:
        return (
            np.isclose(self.min_x, other.min_x)
            and np.isclose(self.min_y, other.min_y)
            and np.isclose(self.min_z, other.min_z)
            and np.isclose(self.max_x, other.max_x)
            and np.isclose(self.max_y, other.max_y)
            and np.isclose(self.max_z, other.max_z)
            and np.allclose(self.origin.to_np(), other.origin.to_np())
        )



import os
from dataclasses import dataclass

from ..world_description.geometry import FileMesh
from ..datastructures.prefixed_name import PrefixedName
from ..spatial_types import TransformationMatrix
from ..world import World
from ..world_description.shape_collection import ShapeCollection
from ..world_description.world_entity import Body


@dataclass
class MeshParser:
    """
    Adapter for mesh files.
    """

    file_path: str
    """
    The path to the mesh file.
    """

    def parse(self) -> World:
        """
        Parse the mesh file to a body and return a world containing that body.

        :return: A World object containing the parsed body.
        """
        file_name = os.path.basename(self.file_path)

        mesh_shape = FileMesh(origin=TransformationMatrix(), filename=self.file_path)
        body = Body(
            name=PrefixedName(file_name),
            collision=ShapeCollection([mesh_shape]),
            visual=ShapeCollection([mesh_shape]),
        )

        world = World()
        with world.modify_world():
            world.add_kinematic_structure_entity(body)

        return world


@dataclass
class STLParser(MeshParser):
    pass


@dataclass
class OBJParser(MeshParser):
    pass


@dataclass
class DAEParser(MeshParser):
    pass


@dataclass
class PLYParser(MeshParser):
    pass


@dataclass
class OFFParser(MeshParser):
    pass


@dataclass
class GLBParser(MeshParser):
    pass


@dataclass
class XYZParser(MeshParser):
    pass



from __future__ import annotations

import itertools
from typing import Optional, List

import trimesh.sample
from krrood.entity_query_language.entity import (
    symbolic_mode,
    let,
    an,
    entity,
    and_,
    not_,
    contains,
    the,
)

from ..collision_checking.collision_detector import Collision, CollisionCheck
from ..collision_checking.trimesh_collision_detector import TrimeshCollisionDetector
from ..robots.abstract_robot import AbstractRobot, ParallelGripper
from ..spatial_computations.raytracer import RayTracer
from ..spatial_types import TransformationMatrix
from ..world_description.world_entity import Body


def robot_in_collision(
    robot: AbstractRobot,
    ignore_collision_with: Optional[List[Body]] = None,
    threshold: float = 0.001,
) -> List[Collision]:
    """
    Check if the robot collides with any object in the world at the given pose.

    :param robot: The robot object
    :param ignore_collision_with: A list of objects to ignore collision with
    :param threshold: The threshold for contact detection
    :return: True if the robot collides with any object, False otherwise
    """

    if ignore_collision_with is None:
        ignore_collision_with = []

    with symbolic_mode():
        body = let(type_=Body, domain=robot._world.bodies_with_enabled_collision)
        possible_collisions_bodies = an(
            entity(
                body,
                and_(
                    not_(contains(robot.bodies, body)),
                    not_(contains(ignore_collision_with, body)),
                ),
            ),
        )
    possible_collisions_bodies = possible_collisions_bodies.evaluate()

    tcd = TrimeshCollisionDetector(robot._world)

    collisions = tcd.check_collisions(
        {
            CollisionCheck(robot_body, collision_body, threshold, robot._world)
            for robot_body, collision_body in itertools.product(
                robot.bodies_with_collisions, possible_collisions_bodies
            )
        }
    )
    return collisions


def robot_holds_body(robot: AbstractRobot, body: Body) -> bool:
    """
    Check if a robot is holding an object.

    :param robot: The robot object
    :param body: The body to check if it is picked
    :return: True if the robot is holding the object, False otherwise
    """
    with symbolic_mode():
        grippers = an(
            entity(
                g := let(ParallelGripper, robot._world.semantic_annotations),
                g._robot == robot,
            )
        )

    return any(
        [is_body_in_gripper(body, gripper) > 0.0 for gripper in grippers.evaluate()]
    )


def blocking(
    pose: TransformationMatrix,
    root: Body,
    tip: Body,
) -> List[Collision]:
    """
    Get the bodies that are blocking the robot from reaching a given position.
    The blocking are all bodies that are in collision with the robot when reaching for the pose.

    :param pose: The pose to reach
    :param root: The root of the kinematic chain.
    :param tip: The threshold between the end effector and the position.
    :return: A list of bodies the robot is in collision with when reaching for the specified object or None if the pose or object is not reachable.
    """
    result = root._world.compute_inverse_kinematics(
        root=root, tip=tip, target=pose, max_iterations=1000
    )
    with root._world.modify_world():
        for dof, state in result.items():
            root._world.state[dof.name].position = state

    with symbolic_mode():
        robot = the(
            entity(
                r := let(AbstractRobot, root._world.semantic_annotations),
                tip in r.bodies,
            )
        )
    return robot_in_collision(robot.evaluate(), [])


def is_body_in_gripper(
    body: Body, gripper: ParallelGripper, sample_size: int = 100
) -> float:
    """
    Check if the body in the gripper.

    This method samples random rays between the finger and the thumb and returns the marginal probability that the rays
    intersect.

    :param body: The body for which the check should be done.
    :param gripper: The gripper for which the check should be done.
    :param sample_size: The number of rays to sample.

    :return: The percentage of rays between the fingers that hit the body.
    """

    # Retrieve meshes in local frames
    thumb_mesh = gripper.thumb.tip.collision.combined_mesh.copy()
    finger_mesh = gripper.finger.tip.collision.combined_mesh.copy()
    body_mesh = body.collision.combined_mesh.copy()

    # Transform copies of the meshes into the world frame
    body_mesh.apply_transform(body.global_pose.to_np())
    thumb_mesh.apply_transform(gripper.thumb.tip.global_pose.to_np())
    finger_mesh.apply_transform(gripper.finger.tip.global_pose.to_np())

    # get random points from thumb mesh
    finger_points = trimesh.sample.sample_surface(finger_mesh, sample_size)[0]
    thumb_points = trimesh.sample.sample_surface(thumb_mesh, sample_size)[0]

    rt = RayTracer(gripper._world)
    rt.update_scene()

    points, index_ray, bodies = rt.ray_test(finger_points, thumb_points)
    return len([b for b in bodies if b == body]) / sample_size



import os
import threading
import time

from krrood.entity_query_language.symbol_graph import SymbolGraph
from krrood.entity_query_language.symbolic import Variable
from typing_extensions import Tuple

import pytest

from .adapters.urdf import URDFParser
from .utils import rclpy_installed, tracy_installed, hsrb_installed
from .world_description.connections import (
    Connection6DoF,
    PrismaticConnection,
    RevoluteConnection,
    FixedConnection,
    OmniDrive,
)
from .world_description.degree_of_freedom import DegreeOfFreedom
from .world_description.geometry import Box, Scale, Sphere
from .datastructures.prefixed_name import PrefixedName
from .spatial_types import TransformationMatrix
from .spatial_types.derivatives import DerivativeMap
from .spatial_types.spatial_types import Vector3
from .world import World
from .world_description.shape_collection import ShapeCollection
from .world_description.world_entity import KinematicStructureEntity, Body


@pytest.fixture
def world_setup() -> Tuple[
    World,
    Body,
    Body,
    Body,
    Body,
    Body,
]:
    world = World()
    root = Body(name=PrefixedName(name="root", prefix="world"))
    l1 = Body(name=PrefixedName("l1"))
    l2 = Body(name=PrefixedName("l2"))
    bf = Body(name=PrefixedName("bf"))
    r1 = Body(name=PrefixedName("r1"))
    r2 = Body(name=PrefixedName("r2"))

    with world.modify_world():
        [world.add_kinematic_structure_entity(b) for b in [root, l1, l2, bf, r1, r2]]
        lower_limits = DerivativeMap()
        lower_limits.velocity = -1
        upper_limits = DerivativeMap()
        upper_limits.velocity = 1
        dof = DegreeOfFreedom(
            name=PrefixedName("dof"),
            lower_limits=lower_limits,
            upper_limits=upper_limits,
        )
        world.add_degree_of_freedom(dof)

        c_l1_l2 = PrismaticConnection(
            parent=l1, child=l2, dof_name=dof.name, axis=Vector3.X(reference_frame=l1)
        )
        c_r1_r2 = RevoluteConnection(
            parent=r1, child=r2, dof_name=dof.name, axis=Vector3.Z(reference_frame=r1)
        )
        bf_root_l1 = FixedConnection(parent=bf, child=l1)
        bf_root_r1 = FixedConnection(parent=bf, child=r1)
        world.add_connection(c_l1_l2)
        world.add_connection(c_r1_r2)
        world.add_connection(bf_root_l1)
        world.add_connection(bf_root_r1)
        c_root_bf = Connection6DoF.create_with_dofs(parent=root, child=bf, world=world)
        world.add_connection(c_root_bf)

    return world, l1, l2, bf, r1, r2


@pytest.fixture
def world_setup_simple():
    world = World()
    root = Body(name=PrefixedName(name="root", prefix="world"))
    body1 = Body(
        name=PrefixedName("name1", prefix="test"),
        collision=ShapeCollection(
            [
                Box(
                    origin=TransformationMatrix.from_xyz_rpy(),
                    scale=Scale(0.25, 0.25, 0.25),
                )
            ]
        ),
    )
    body2 = Body(
        name=PrefixedName("name2", prefix="test"),
        collision=ShapeCollection(
            [
                Box(
                    origin=TransformationMatrix.from_xyz_rpy(),
                    scale=Scale(0.25, 0.25, 0.25),
                )
            ]
        ),
    )
    body3 = Body(
        name=PrefixedName("name3", prefix="test"),
        collision=ShapeCollection(
            [Sphere(origin=TransformationMatrix.from_xyz_rpy(), radius=0.01)]
        ),
    )

    body4 = Body(
        name=PrefixedName("name4", prefix="test"),
        collision=ShapeCollection(
            [Sphere(origin=TransformationMatrix.from_xyz_rpy(), radius=0.01)]
        ),
    )

    with world.modify_world():
        world.add_kinematic_structure_entity(body1)
        world.add_kinematic_structure_entity(body2)
        world.add_kinematic_structure_entity(body3)
        world.add_kinematic_structure_entity(body4)

        c_root_body1 = Connection6DoF.create_with_dofs(
            parent=root, child=body1, world=world
        )
        c_root_body2 = Connection6DoF.create_with_dofs(
            parent=root, child=body2, world=world
        )
        c_root_body3 = Connection6DoF.create_with_dofs(
            parent=root, child=body3, world=world
        )
        c_root_body4 = Connection6DoF.create_with_dofs(
            parent=root, child=body4, world=world
        )

        world.add_connection(c_root_body1)
        world.add_connection(c_root_body2)
        world.add_connection(c_root_body3)
        world.add_connection(c_root_body4)
    return world, body1, body2, body3, body4


@pytest.fixture
def two_arm_robot_world():
    urdf_dir = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "..", "..", "resources", "urdf"
    )
    robot = os.path.join(urdf_dir, "simple_two_arm_robot.urdf")
    world = World()
    with world.modify_world():
        localization_body = Body(name=PrefixedName("odom_combined"))
        world.add_kinematic_structure_entity(localization_body)

        robot_parser = URDFParser.from_file(file_path=robot)
        world_with_robot = robot_parser.parse()
        # world_with_pr2.plot_kinematic_structure()
        root = world_with_robot.root
        c_root_bf = OmniDrive.create_with_dofs(
            parent=localization_body, child=root, world=world
        )
        world.merge_world(world_with_robot, root_connection=c_root_bf)
    return world


@pytest.fixture
def pr2_world():
    urdf_dir = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "..", "..", "resources", "urdf"
    )
    pr2 = os.path.join(urdf_dir, "pr2_kinematic_tree.urdf")
    pr2_parser = URDFParser.from_file(file_path=pr2)
    world_with_pr2 = pr2_parser.parse()
    with world_with_pr2.modify_world():
        pr2_root = world_with_pr2.root
        localization_body = Body(name=PrefixedName("odom_combined"))
        world_with_pr2.add_kinematic_structure_entity(localization_body)
        c_root_bf = OmniDrive.create_with_dofs(
            parent=localization_body, child=pr2_root, world=world_with_pr2
        )
        world_with_pr2.add_connection(c_root_bf)

    return world_with_pr2


@pytest.fixture
def tracy_world():
    if not tracy_installed():
        pytest.skip("Tracy not installed")
    urdf_dir = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "..", "..", "resources", "urdf"
    )
    tracy = os.path.join(urdf_dir, "tracy.urdf")
    world = World()
    with world.modify_world():
        localization_body = Body(name=PrefixedName("odom_combined"))
        world.add_kinematic_structure_entity(localization_body)

        tracy_parser = URDFParser.from_file(file_path=tracy)
        world_with_tracy = tracy_parser.parse()
        # world_with_tracy.plot_kinematic_structure()
        tracy_root = world_with_tracy.root
        c_root_bf = Connection6DoF.create_with_dofs(
            parent=localization_body, child=tracy_root, world=world
        )
        world.merge_world(world_with_tracy, c_root_bf)

    return world


@pytest.fixture
def hsrb_world():
    if not hsrb_installed():
        pytest.skip("HSRB not installed")
    urdf_dir = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "..", "..", "resources", "urdf"
    )
    hsrb = os.path.join(urdf_dir, "hsrb.urdf")
    world = World()
    with world.modify_world():
        localization_body = Body(name=PrefixedName("odom_combined"))
        world.add_kinematic_structure_entity(localization_body)

        hsrb_parser = URDFParser.from_file(file_path=hsrb)
        world_with_hsrb = hsrb_parser.parse()
        hsrb_root = world_with_hsrb.root
        c_root_bf = Connection6DoF.create_with_dofs(
            parent=localization_body, child=hsrb_root, world=world
        )
        world.merge_world(world_with_hsrb, c_root_bf)

    return world


@pytest.fixture
def apartment_world() -> World:
    """
    Return the apartment world parsed from the URDF file.
    """
    urdf_dir = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "..", "..", "resources", "urdf"
    )
    apartment = os.path.join(urdf_dir, "apartment.urdf")
    parser = URDFParser.from_file(file_path=apartment)
    world = parser.parse()
    world.validate()
    return world


@pytest.fixture(scope="function")
def rclpy_node():
    if not rclpy_installed():
        pytest.skip("ROS not installed")
    import rclpy
    from rclpy.executors import SingleThreadedExecutor

    rclpy.init()
    node = rclpy.create_node("test_node")

    executor = SingleThreadedExecutor()
    executor.add_node(node)

    thread = threading.Thread(target=executor.spin, daemon=True, name="rclpy-executor")
    thread.start()
    time.sleep(0.1)
    try:
        yield node
    finally:
        # Stop executor cleanly and wait for the thread to exit
        executor.shutdown()
        thread.join(timeout=2.0)

        # Remove the node from the executor and destroy it
        # (executor.shutdown() takes care of spinning; add_node is safe to keep as-is)
        node.destroy_node()

        # Shut down the ROS client library
        rclpy.shutdown()


@pytest.fixture(autouse=True, scope="function")
def cleanup_after_test():
    # Setup: runs before each test
    yield
    # Teardown: runs after each test
    SymbolGraph().clear()

@pytest.fixture()
def kitchen_world():
    path = os.path.join(
        os.path.dirname(__file__), "..", "..", "resources", "urdf", "kitchen-small.urdf"
    )
    parser = URDFParser.from_file(file_path=path)
    world = parser.parse()
    world.validate()
    return world



import os
import unittest

from semantic_digital_twin.adapters.mesh import STLParser
from semantic_digital_twin.world_description.geometry import FileMesh


class STLAdapterTestCase(unittest.TestCase):

    def setUp(self):
        # Set up any necessary resources or state before each test
        self.milk_path = os.path.join(
            os.path.dirname(__file__), "..", "..", "resources", "stl", "milk.stl"
        )
        self.cup = os.path.join(
            os.path.dirname(__file__), "..", "..", "resources", "stl", "jeroen_cup.stl"
        )

    def test_stl_parsing_construct(self):
        stl_parser = STLParser(self.milk_path)
        self.assertEqual(stl_parser.file_path, self.milk_path)

    def test_stl_parsing(self):
        stl_parser = STLParser(self.milk_path)
        world = stl_parser.parse()
        world.validate()

        self.assertEqual(len(world.kinematic_structure_entities), 1)
        self.assertEqual(len(world.kinematic_structure_entities[0].collision), 1)
        self.assertEqual(len(world.kinematic_structure_entities[0].visual), 1)
        self.assertEqual(
            FileMesh, type(world.kinematic_structure_entities[0].collision[0])
        )

    def test_parse_and_merge(self):
        milk_world = STLParser(self.milk_path).parse()
        cup_world = STLParser(self.cup).parse()

        milk_world.merge_world(cup_world)
        self.assertEqual(len(milk_world.kinematic_structure_entities), 2)
        self.assertEqual(len(milk_world.kinematic_structure_entities[0].collision), 1)



import os
import unittest

import plotly.graph_objects as go
from random_events.interval import SimpleInterval
from random_events.product_algebra import SimpleEvent

from semantic_digital_twin.adapters.urdf import URDFParser
from semantic_digital_twin.world_description.geometry import BoundingBox
from semantic_digital_twin.world_description.shape_collection import (
    BoundingBoxCollection,
    ShapeCollection,
)
from semantic_digital_twin.world_description.graph_of_convex_sets import (
    GraphOfConvexSets,
    PoseOccupiedError,
)
from semantic_digital_twin.spatial_types import Point3
from semantic_digital_twin.spatial_types.spatial_types import TransformationMatrix
from semantic_digital_twin.datastructures.variables import SpatialVariables
from semantic_digital_twin.world import World
from semantic_digital_twin.world_description.world_entity import Body


class GCSTestCase(unittest.TestCase):
    """
    Testcase to test the navigation around a unit box.
    """

    gcs: GraphOfConvexSets

    @classmethod
    def setUpClass(cls):
        world = World()
        with world.modify_world():
            world.add_kinematic_structure_entity(Body())
        gcs = GraphOfConvexSets(world)

        obstacle = BoundingBox(0, 0, 0, 1, 1, 1, world.root)

        z_lim = SimpleInterval(0.45, 0.55)
        x_lim = SimpleInterval(-2, 3)
        y_lim = SimpleInterval(-2, 3)
        limiting_event = SimpleEvent(
            {
                SpatialVariables.x.value: x_lim,
                SpatialVariables.y.value: y_lim,
                SpatialVariables.z.value: z_lim,
            }
        )
        obstacles = BoundingBoxCollection.from_event(
            world.root,
            ~obstacle.simple_event.as_composite_set()
            & limiting_event.as_composite_set(),
        )
        [gcs.add_node(bb) for bb in obstacles]
        gcs.calculate_connectivity()
        cls.gcs = gcs

    def test_reachability(self):
        start_point = Point3(-1, -1, 0.5)
        target_point = Point3(2, 2, 0.5)

        path = self.gcs.path_from_to(start_point, target_point)
        self.assertEqual(len(path), 4)

    def test_plot(self):
        free_space_plot = go.Figure(self.gcs.plot_free_space())
        self.assertIsNotNone(free_space_plot)
        occupied_space_plot = go.Figure(self.gcs.plot_occupied_space())
        self.assertIsNotNone(occupied_space_plot)


class GCSFromWorldTestCase(unittest.TestCase):
    """
    Test the application of a connectivity graph to the belief state.
    """

    world: World

    @classmethod
    def setUpClass(cls):
        urdf_dir = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "..", "..", "resources", "urdf"
        )
        apartment = os.path.join(urdf_dir, "table.urdf")
        apartment_parser = URDFParser.from_file(file_path=apartment)
        cls.world = apartment_parser.parse()

    def test_from_world(self):
        search_space = BoundingBoxCollection(
            [
                BoundingBox(
                    min_x=-5,
                    max_x=-2,
                    min_y=-1,
                    max_y=2,
                    min_z=0,
                    max_z=2,
                    origin=TransformationMatrix(reference_frame=self.world.root),
                )
            ],
            self.world.root,
        )
        gcs = GraphOfConvexSets.free_space_from_world(
            self.world, search_space=search_space
        )
        self.assertIsNotNone(gcs)
        self.assertGreater(len(gcs.graph.nodes()), 0)
        self.assertGreater(len(gcs.graph.edges()), 0)

        start = Point3(-4.5, -0.5, 0.4)
        target = Point3(-2.5, 1.5, 0.9)

        path = gcs.path_from_to(start, target)

        self.assertIsNotNone(path)
        self.assertGreater(len(path), 1)

        with self.assertRaises(PoseOccupiedError):
            start = Point3(-10, -10, -10)
            target = Point3(10, 10, 10)
            gcs.path_from_to(start, target)

    def test_navigation_map_from_world(self):
        search_space = BoundingBoxCollection(
            [
                BoundingBox(
                    min_x=-5,
                    max_x=-2,
                    min_y=-1,
                    max_y=2,
                    min_z=0,
                    max_z=2,
                    origin=TransformationMatrix(reference_frame=self.world.root),
                )
            ],
            self.world.root,
        )
        gcs = GraphOfConvexSets.navigation_map_from_world(
            self.world, search_space=search_space
        )
        self.assertGreater(len(gcs.graph.nodes()), 0)
        self.assertGreater(len(gcs.graph.edges()), 0)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass, field
from enum import IntEnum
from typing_extensions import Generic, TypeVar, List, Optional, Dict, Any

from random_events.utils import SubclassJSONSerializer

T = TypeVar("T")


class Derivatives(IntEnum):
    """
    Enumaration of interpretation for the order of derivativeson the spatial positions
    """

    position = 0
    velocity = 1
    acceleration = 2
    jerk = 3
    snap = 4
    crackle = 5
    pop = 6

    @classmethod
    def range(cls, start: Derivatives, stop: Derivatives, step: int = 1):
        """
        Includes stop!
        """
        return [item for item in cls if start <= item <= stop][::step]


@dataclass
class DerivativeMap(Generic[T], SubclassJSONSerializer):
    """
    A container class that maps derivatives (position, velocity, acceleration, jerk) to values of type T.

    This class provides a structured way to store and access different orders of derivatives
    using properties. Each derivative order can hold a value of type T or None.

    Type Parameters:
        T: The type of values stored for each derivative order.

    Attributes:
        data (List[Optional[T]]): Internal list storing the derivative values, initialized with None values.
    """

    data: List[Optional[T]] = field(default_factory=lambda: [None] * len(Derivatives))
    """
    Internal list storing the derivative values, initialized with None values.
    Order corresponds to the order of the Derivatives enum.
    """

    def __hash__(self):
        return hash(tuple(self.data))

    @property
    def position(self) -> T:
        return self.data[Derivatives.position]

    @position.setter
    def position(self, value: T):
        self.data[Derivatives.position] = value

    @property
    def velocity(self) -> T:
        return self.data[Derivatives.velocity]

    @velocity.setter
    def velocity(self, value: T):
        self.data[Derivatives.velocity] = value

    @property
    def acceleration(self) -> T:
        return self.data[Derivatives.acceleration]

    @acceleration.setter
    def acceleration(self, value: T):
        self.data[Derivatives.acceleration] = value

    @property
    def jerk(self) -> T:
        return self.data[Derivatives.jerk]

    @jerk.setter
    def jerk(self, value: T):
        self.data[Derivatives.jerk] = value

    def to_json(self) -> Dict[str, Any]:
        return {**super().to_json(), "data": self.data}

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> DerivativeMap[T]:
        return cls(data=data["data"])

    def __mul__(self, other: float) -> DerivativeMap[T]:
        result = DerivativeMap()
        for i, v in enumerate(self.data):
            if v is not None:
                result.data[i] = v * other
        return result

    def __add__(self, other: float) -> DerivativeMap[T]:
        result = DerivativeMap()
        for i, v in enumerate(self.data):
            if v is not None:
                result.data[i] = v + other
        return result



from __future__ import annotations

import itertools
import logging
from dataclasses import dataclass, field
from functools import cached_property
from typing_extensions import Dict, Any, Self, Optional, List, Iterator

import numpy as np
from random_events.product_algebra import Event, SimpleEvent
from random_events.utils import SubclassJSONSerializer
from trimesh import Trimesh
from trimesh.util import concatenate
from typing_extensions import TYPE_CHECKING

from .geometry import Shape, BoundingBox
from ..datastructures.variables import SpatialVariables
from ..spatial_types import TransformationMatrix, Point3

if TYPE_CHECKING:
    from .world_entity import KinematicStructureEntity
    from ..world import World

logger = logging.getLogger(__name__)


@dataclass
class ShapeCollection(SubclassJSONSerializer):
    """
    A collection of shapes.
    """

    shapes: List[Shape] = field(default_factory=list)
    """
    The shapes contained in this collection.
    """

    reference_frame: Optional[KinematicStructureEntity] = None
    """
    Backreference to the kinematic structure entity this collection belongs to.
    """

    @property
    def world(self) -> Optional[World]:
        """
        The world this collection belongs to.
        """
        if self.reference_frame is not None:
            return self.reference_frame._world
        return None

    def transform_all_shapes_to_own_frame(self):
        """
        Transform all shapes into this collections' frame in-place.
        """
        if self.reference_frame is None:
            return
        for shape in self.shapes:
            self._transform_to_own_frame(shape)

    def _transform_to_own_frame(self, shape: Shape):
        """
        Transform the shape to this collections' frame in-place.
        :param shape: The shape to transform.
        """
        if shape.origin.reference_frame is None:
            # If we don’t have a world, fall back to the owning body/frame
            shape.origin.reference_frame = self.reference_frame
        elif (
                self.reference_frame is not None
                and shape.origin.reference_frame != self.reference_frame
                and self.reference_frame._world is not None
        ):
            logger.warning(
                f"Transformed shape {shape} to {self.reference_frame} since it was in a different "
                f"reference frame than the collection."
            )
            shape.origin = self.reference_frame._world.transform(
                shape.origin.reference_frame.global_pose,
                self.reference_frame,
            )

    def __getitem__(self, index: int) -> Shape:
        return self.shapes[index]

    def __len__(self) -> int:
        return len(self.shapes)

    def __iter__(self) -> Iterator[Shape]:
        return iter(self.shapes)

    def __contains__(self, shape: Shape) -> bool:
        return shape in self.shapes

    def append(self, shape: Shape):
        if self.world is not None:
            self._transform_to_own_frame(
                shape,
            )
        self.shapes.append(shape)

    @cached_property
    def combined_mesh(self) -> Trimesh:
        """
        Combines all shapes into a single mesh, applying the respective transformations.
        :return: A single Trimesh representing the combined collision geometry.
        """
        transformed_meshes = []
        for shape in self.shapes:
            transform = shape.origin.to_np()
            mesh = shape.mesh.copy()
            mesh.apply_transform(transform)
            transformed_meshes.append(mesh)
        return concatenate(transformed_meshes)

    def as_bounding_box_collection_at_origin(
            self, origin: TransformationMatrix
    ) -> BoundingBoxCollection:
        """
        Provides the bounding box collection for this entity given a transformation matrix as origin.
        :param origin: The origin to express the bounding boxes from.
        :returns: A collection of bounding boxes in world-space coordinates.
        """
        world_bboxes = []

        for shape in self.shapes:
            if shape.origin.reference_frame is None:
                continue
            local_bb: BoundingBox = shape.local_frame_bounding_box
            world_bb = local_bb.transform_to_origin(origin)
            world_bboxes.append(world_bb)

        return BoundingBoxCollection(
            world_bboxes,
            origin.reference_frame,
        )

    def as_bounding_box_collection_in_frame(
            self, reference_frame: KinematicStructureEntity
    ) -> BoundingBoxCollection:
        """
        Provides the bounding box collection for this entity in the given reference frame.
        :param reference_frame: The reference frame to express the bounding boxes in.
        :returns: A collection of bounding boxes in world-space coordinates.
        """
        return self.as_bounding_box_collection_at_origin(
            TransformationMatrix(reference_frame=reference_frame)
        )

    def to_json(self) -> Dict[str, Any]:
        return {
            **super().to_json(),
            "shapes": [shape.to_json() for shape in self.shapes],
        }

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        return cls(shapes=[Shape.from_json(d) for d in data["shapes"]])

    def center_of_mass_in_world(self) -> Point3:
        """
        :return: The center of mass of this shape collection in the world coordinate frame.
        """
        # Center of mass in the body's local frame (collision geometry)
        com_local: np.ndarray[np.float64] = self.combined_mesh.center_mass  # (3,)
        # Transform to world frame using the body's global pose
        com = Point3(
            x_init=com_local[0],
            y_init=com_local[1],
            z_init=com_local[2],
            reference_frame=self.reference_frame,
        )
        return self.world.transform(com, self.world.root)


@dataclass
class BoundingBoxCollection(ShapeCollection):
    """
    Dataclass for storing a collection of bounding boxes.
    """

    shapes: List[BoundingBox]

    def __post_init__(self):
        if not self.reference_frame:
            raise ValueError("BoundingBoxCollection must have a reference frame.")
        for box in self.bounding_boxes:
            assert (
                    box.origin.reference_frame == self.reference_frame
            ), "All bounding boxes must have the same reference frame."

    def __iter__(self) -> Iterator[BoundingBox]:
        return iter(self.bounding_boxes)

    @property
    def bounding_boxes(self) -> List[BoundingBox]:
        return self.shapes

    @property
    def event(self) -> Event:
        """
        :return: The bounding boxes as a random event.
        """
        return Event(*[box.simple_event for box in self.bounding_boxes])

    def merge(self, other: BoundingBoxCollection) -> BoundingBoxCollection:
        """
        Merge another bounding box collection into this one.

        :param other: The other bounding box collection.
        :return: The merged bounding box collection.
        """
        assert (
                self.reference_frame == other.reference_frame
        ), "The reference frames of the bounding box collections must be the same."
        return BoundingBoxCollection(
            reference_frame=self.reference_frame,
            shapes=self.bounding_boxes + other.bounding_boxes,
        )

    def bloat(
            self, x_amount: float = 0.0, y_amount: float = 0, z_amount: float = 0
    ) -> BoundingBoxCollection:
        """
        Enlarges all bounding boxes in the collection by a given amount in all dimensions.

        :param x_amount: The amount to adjust the x-coordinates
        :param y_amount: The amount to adjust the y-coordinates
        :param z_amount: The amount to adjust the z-coordinates

        :return: The enlarged bounding box collection
        """
        return BoundingBoxCollection(
            [box.bloat(x_amount, y_amount, z_amount) for box in self.bounding_boxes],
            self.reference_frame,
        )

    @classmethod
    def from_simple_event(
            cls,
            reference_frame: KinematicStructureEntity,
            simple_event: SimpleEvent,
            keep_surface: bool = False,
    ) -> BoundingBoxCollection:
        """
        Create a list of bounding boxes from a simple random event.

        :param reference_frame: The reference frame of the bounding boxes.
        :param simple_event: The random event.
        :param keep_surface: Whether to keep events that are infinitely thin
        :return: The list of bounding boxes.
        """
        result = []
        for x, y, z in itertools.product(
                simple_event[SpatialVariables.x.value].simple_sets,
                simple_event[SpatialVariables.y.value].simple_sets,
                simple_event[SpatialVariables.z.value].simple_sets,
        ):

            bb = BoundingBox(
                x.lower,
                y.lower,
                z.lower,
                x.upper,
                y.upper,
                z.upper,
                TransformationMatrix.from_xyz_quaternion(x.upper - (x.upper - x.lower) / 2, y.upper - (y.upper - y.lower) / 2,
                                                         z.upper - (z.upper - z.lower) / 2, 0, 0, 0, 1,
                                                         reference_frame=reference_frame)
            )
            if not keep_surface and (bb.depth == 0 or bb.height == 0 or bb.width == 0):
                continue
            result.append(bb)
        return BoundingBoxCollection(result, reference_frame)

    @classmethod
    def from_event(
            cls, reference_frame: KinematicStructureEntity, event: Event
    ) -> Self:
        """
        Create a list of bounding boxes from a random event.

        :param reference_frame: The reference frame of the bounding boxes.
        :param event: The random event.
        :return: The list of bounding boxes.
        """
        return cls(
            [
                box
                for simple_event in event.simple_sets
                for box in cls.from_simple_event(reference_frame, simple_event)
            ],
            reference_frame,
        )

    @classmethod
    def from_shapes(cls, shapes: ShapeCollection) -> Self:
        """
        Create a bounding box collection from a list of shapes.

        :param shapes: The list of shapes.
        :return: The bounding box collection.
        """
        if len(shapes) == 0:
            return cls(shapes=[])
        for shape in shapes:
            assert (
                    shape.origin.reference_frame == shapes[0].origin.reference_frame
            ), "All shapes must have the same reference frame."

        local_bbs = [shape.local_frame_bounding_box for shape in shapes]
        reference_frame = shapes[0].origin.reference_frame
        return cls(
            [bb.transform_to_origin(bb.origin) for bb in local_bbs],
            reference_frame,
        )

    def as_shapes(self) -> ShapeCollection:
        return ShapeCollection(
            [box.as_shape() for box in self.bounding_boxes],
            self.reference_frame,
        )

    def bounding_box(self) -> BoundingBox:
        """
        Get the 8 corners of a bounding box that contains all bounding boxes in the collection.

        :return: A list of Point3 objects representing the corners of the bounding box.
        """
        all_x = [bb.min_x for bb in self.bounding_boxes] + [
            bb.max_x for bb in self.bounding_boxes
        ]
        all_y = [bb.min_y for bb in self.bounding_boxes] + [
            bb.max_y for bb in self.bounding_boxes
        ]
        all_z = [bb.min_z for bb in self.bounding_boxes] + [
            bb.max_z for bb in self.bounding_boxes
        ]
        return BoundingBox(
            min(all_x),
            min(all_y),
            min(all_z),
            max(all_x),
            max(all_y),
            max(all_z),
            TransformationMatrix.from_xyz_quaternion(reference_frame=self.reference_frame),
        )



from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass, field
from typing_extensions import Dict, Any

from random_events.utils import SubclassJSONSerializer

from ..datastructures.prefixed_name import PrefixedName
from ..exceptions import UsageError
from ..spatial_types import spatial_types as cas
from ..spatial_types.derivatives import Derivatives, DerivativeMap
from ..spatial_types.symbol_manager import symbol_manager
from .world_entity import WorldEntity


@dataclass
class DegreeOfFreedom(WorldEntity, SubclassJSONSerializer):
    """
    A class representing a degree of freedom in a world model with associated derivatives and limits.

    This class manages a variable that can freely change within specified limits, tracking its position,
    velocity, acceleration, and jerk. It maintains symbolic representations for each derivative order
    and provides methods to get and set limits for these derivatives.
    """

    lower_limits: DerivativeMap[float] = field(default_factory=DerivativeMap)
    upper_limits: DerivativeMap[float] = field(default_factory=DerivativeMap)
    """
    Lower and upper bounds for each derivative
    """

    symbols: DerivativeMap[cas.Symbol] = field(
        default_factory=DerivativeMap, init=False
    )
    """
    Symbolic representations for each derivative
    """

    has_hardware_interface: bool = False
    """
    Whether this DOF is linked to a controller and can therefore respond to control commands.

    E.g. the caster wheels of a PR2 have dofs, but they are not directly controlled. 
    Instead a the omni drive connection is directly controlled and a low level controller translates these commands
    to commands for the caster wheels.

    A door hinge also has a dof that cannot be controlled.
    """

    def __post_init__(self):
        self.lower_limits = self.lower_limits or DerivativeMap()
        self.upper_limits = self.upper_limits or DerivativeMap()

    def create_and_register_symbols(self):
        assert self._world is not None
        for derivative in Derivatives.range(Derivatives.position, Derivatives.jerk):
            s = symbol_manager.register_symbol_provider(
                f"{self.name}_{derivative}",
                lambda d=derivative: self._world.state[self.name][d],
            )
            self.symbols.data[derivative] = s

    def has_position_limits(self) -> bool:
        try:
            lower_limit = self.lower_limits.position
            upper_limit = self.upper_limits.position
            return lower_limit is not None or upper_limit is not None
        except KeyError:
            return False

    def __hash__(self):
        return hash(id(self))

    def to_json(self) -> Dict[str, Any]:
        return {
            **super().to_json(),
            "lower_limits": self.lower_limits.to_json(),
            "upper_limits": self.upper_limits.to_json(),
            "name": self.name.to_json(),
        }

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> DegreeOfFreedom:
        lower_limits = DerivativeMap.from_json(data["lower_limits"])
        upper_limits = DerivativeMap.from_json(data["upper_limits"])
        return cls(
            name=PrefixedName.from_json(data["name"]),
            lower_limits=lower_limits,
            upper_limits=upper_limits,
        )

    def __deepcopy__(self, memo):
        result = DegreeOfFreedom(
            lower_limits=deepcopy(self.lower_limits),
            upper_limits=deepcopy(self.upper_limits),
            name=self.name,
            has_hardware_interface=self.has_hardware_interface,
        )
        result._world = self._world
        # there can't be two symbols with the same name anyway
        result.symbols = self.symbols
        return result

    def _overwrite_dof_limits(
        self,
        new_lower_limits: DerivativeMap[float],
        new_upper_limits: DerivativeMap[float],
    ):
        """
        Overwrites the degree-of-freedom (DOF) limits for a range of derivatives. This updates
        lower and upper limits based on the given new limits. For each derivative, if the
        new limit is provided and it is more restrictive than the original limit, the limit
        will be updated accordingly.

        :param new_lower_limits: A mapping of new lower limits for the specified derivatives.
            If a new lower limit is None, no change is applied for that derivative.
        :param new_upper_limits: A mapping of new upper limits for the specified derivatives.
            If a new upper limit is None, no change is applied for that derivative.
        """
        if not isinstance(self.symbols.position, cas.Symbol):
            raise UsageError(
                "Cannot overwrite limits of mimic DOFs, use .raw_dof._overwrite_dof_limits instead."
            )
        for derivative in Derivatives.range(Derivatives.position, Derivatives.jerk):
            if new_lower_limits.data[derivative] is not None:
                if self.lower_limits.data[derivative] is None:
                    self.lower_limits.data[derivative] = new_lower_limits.data[
                        derivative
                    ]
                else:
                    self.lower_limits.data[derivative] = max(
                        new_lower_limits.data[derivative],
                        self.lower_limits.data[derivative],
                    )
            if new_upper_limits.data[derivative] is not None:
                if self.upper_limits.data[derivative] is None:
                    self.upper_limits.data[derivative] = new_upper_limits.data[
                        derivative
                    ]
                else:
                    self.upper_limits.data[derivative] = min(
                        new_upper_limits.data[derivative],
                        self.upper_limits.data[derivative],
                    )



import numpy as np

from semantic_digital_twin.spatial_computations.raytracer import RayTracer
from semantic_digital_twin.spatial_types.spatial_types import TransformationMatrix
from semantic_digital_twin.testing import world_setup_simple


def test_create_segmentation_mask(world_setup_simple):
    world, body1, body2, body3, body4 = world_setup_simple
    rt = RayTracer(world)
    rt.update_scene()

    camera_pose = np.array(
        [
            [1.0, 0.0, 0.0, -2.5],
            [0.0, 1.0, 0.0, 0.0],
            [0.0, 0.0, 1.0, 0.2],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )

    world.get_connection(world.root, body2).origin = np.array(
        [[1, 0, 0, -1], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]]
    )

    seg = rt.create_segmentation_mask(
        TransformationMatrix(camera_pose, reference_frame=world.root), resolution=256
    )
    assert seg.shape == (256, 256)  # Assuming a standard resolution

    hit, index, body = rt.ray_test(np.array([1, 0, 1]), np.array([-1, 0, 1]))
    assert hit is not None
    assert index is not None
    assert body is not None
    assert body1.index in seg
    assert body2.index in seg


def test_create_depth_map(world_setup_simple):
    world, body1, body2, body3, body4 = world_setup_simple
    rt = RayTracer(world)
    rt.update_scene()

    camera_pose = np.array(
        [
            [1.0, 0.0, 0.0, -2.5],
            [0.0, 1.0, 0.0, 0.0],
            [0.0, 0.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )

    depth_map = rt.create_depth_map(
        TransformationMatrix(camera_pose, reference_frame=world.root), resolution=512
    )
    assert depth_map is not None
    assert depth_map[0, 0] == -1  # Assuming no objects are hit at the upper left corner
    assert depth_map.shape == (512, 512)
    assert depth_map.max() <= 2.5
    assert depth_map[depth_map > 0].min() >= 2.375


def test_ray_test(world_setup_simple):
    world, body1, body2, body3, body4 = world_setup_simple
    rt = RayTracer(world)
    rt.update_scene()

    hit, index, body = rt.ray_test(np.array([1, 0, 0.1]), np.array([-1, 0, 0.1]))
    assert hit is not None
    assert index is not None
    assert body is not None
    assert body1 in body

    # Test with a ray that does not hit any object
    hit, index, body = rt.ray_test(np.array([10, 10, 10]), np.array([20, 20, 20]))
    assert not np.any(hit)
    assert not np.any(index)
    assert not np.any(body)


def test_ray_test_batch(world_setup_simple):
    world, body1, body2, body3, body4 = world_setup_simple
    world.get_connection(world.root, body1).origin = np.array(
        [[1, 0, 0, 1], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]]
    )
    world.get_connection(world.root, body2).origin = np.array(
        [[1, 0, 0, -1], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]]
    )

    rt = RayTracer(world)
    rt.update_scene()

    rays = np.array([[1, 1, 0.1], [-1, 1, 0.1]])
    targets = np.array([[1, -1, 0.1], [-1, -1, 0.1]])

    hits, indices, bodies = rt.ray_test(rays, targets)
    assert hits is not None
    assert indices is not None
    assert bodies is not None
    assert len(hits) == len(rays)
    assert len(indices) == len(rays)
    assert len(bodies) == len(rays)
    # Test return
    assert bodies[0] == body1
    assert bodies[1] == body2



import numpy as np
import pytest

from semantic_digital_twin.testing import world_setup_simple


def test_closest_points(world_setup_simple):
    world, body1, body2, body3, _ = world_setup_simple

    world.get_connection(world.root, body1).origin = np.array(
        [[1, 0, 0, 1], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]]
    )

    point_body1, point_body2, dist = body1.compute_closest_points_multi([body2])

    # Check that y and z coordinates are the same since we only moved in the x direction
    assert point_body1[0][1] == point_body2[0][1]
    assert point_body1[0][2] == point_body2[0][2]
    # Check that the distance between the points is the distance between the two boxes
    assert dist == pytest.approx(0.75, abs=1e-2)


def test_closest_points_approximate(world_setup_simple):
    world, body1, body2, body3, _ = world_setup_simple

    world.get_connection(world.root, body3).origin = np.array(
        [[1, 0, 0, 0], [0, 1, 0, 0.5], [0, 0, 1, 0.125], [0, 0, 0, 1]]
    )

    point_body3, point_body1, dist = body3.compute_closest_points_multi(
        [body1], sample_size=20
    )

    # Check that the distance between the found points to be within 2 cm
    assert dist == pytest.approx(
        0.385, abs=2e-2
    )  # sphere is 0.385 meter away from the surface of the box plus the radius of the sphere (0.01 m)


def test_closest_points_transform(world_setup_simple):
    world, body1, body2, body3, body4 = world_setup_simple

    world.get_connection(world.root, body4).origin = np.array(
        [[1, 0, 0, 1], [0, 1, 0, 1], [0, 0, 1, 1], [0, 0, 0, 1]]
    )

    point_body3, point_body4, dist = body3.compute_closest_points_multi(
        [body4], sample_size=20
    )

    assert point_body4[0][0] == pytest.approx(
        1.0, abs=1e-2
    )  # 1 meter away from the surface of the box plus the radius of the sphere (0.01 m)
    assert point_body4[0][1] == pytest.approx(1.0, abs=1e-2)
    assert point_body4[0][2] == pytest.approx(
        1.0, abs=1e-2
    )  # 1 meter away from the surface of the box plus the radius of the sphere (0.01 m)


def test_closest_points_multi(world_setup_simple):
    world, body1, body2, body3, body4 = world_setup_simple

    world.get_connection(world.root, body3).origin = np.array(
        [[1, 0, 0, 1], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]]
    )

    world.get_connection(world.root, body4).origin = np.array(
        [[1, 0, 0, -1], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]]
    )

    point_body1, point_other_bodies, dist_1 = body1.compute_closest_points_multi(
        [body3, body4], sample_size=20
    )

    assert dist_1[0] == pytest.approx(
        0.875, abs=2e-2
    )  # sphere is 2.385 meter away from the surface of the box plus the radius of the sphere (0.01 m)
    assert dist_1[1] == pytest.approx(
        0.875, abs=2e-2
    )  # sphere is 0.385 meter away from the surface of the box plus the radius of the sphere (0.01 m)

    assert point_body1[0][0] == pytest.approx(0.125, abs=1e-2)
    assert point_body1[1][0] == pytest.approx(-0.125, abs=1e-2)



import os

from sqlalchemy.exc import OperationalError
from sqlalchemy import create_engine


def persistent_database_available() -> bool:
    """
    Check if a persistent database is available for connection.

    This function validates if the environment variable `semantic_digital_twin_DATABASE_URI`
    is set and attempts to establish a connection with the database using the
    provided URI. If the connection is successful, it returns True, indicating the
    database is available. Otherwise, it returns False.

    :return: Indicates whether the persistent database is accessible
    """
    semantic_digital_twin_database_uri = os.environ.get(
        "semantic_digital_twin_DATABASE_URI"
    )
    if semantic_digital_twin_database_uri is None:
        return False

    try:
        engine = create_engine(f"mysql+pymysql://{semantic_digital_twin_database_uri}")
        with engine.connect():
            ...
    except OperationalError as e:
        return False
    return True



from __future__ import annotations

import os
from collections import defaultdict
from dataclasses import dataclass
from typing import Self

from .robot_mixins import HasNeck, SpecifiesLeftRightArm
from ..datastructures.prefixed_name import PrefixedName
from ..robots.abstract_robot import (
    Neck,
    Finger,
    ParallelGripper,
    Arm,
    Camera,
    FieldOfView,
    Torso,
    AbstractRobot,
)
from ..spatial_types import Quaternion, Vector3
from ..world import World


@dataclass
class PR2(AbstractRobot, SpecifiesLeftRightArm, HasNeck):
    """
    Represents the Personal Robot 2 (PR2), which was originally created by Willow Garage.
    The PR2 robot consists of two arms, each with a parallel gripper, a head with a camera, and a prismatic torso
    """

    def __hash__(self):
        return hash(
            tuple(
                [self.__class__]
                + sorted([kse.name for kse in self.kinematic_structure_entities])
            )
        )

    def load_srdf(self):
        """
        Loads the SRDF file for the PR2 robot, if it exists.
        """
        srdf_path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "..",
            "..",
            "resources",
            "collision_configs",
            "pr2.srdf",
        )
        self._world.load_collision_srdf(srdf_path)

    @classmethod
    def from_world(cls, world: World) -> Self:
        """
        Creates a PR2 robot view from the given world.

        :param world: The world from which to create the robot view.

        :return: A PR2 robot view.
        """

        with world.modify_world():
            robot = cls(
                name=PrefixedName(name="pr2", prefix=world.name),
                root=world.get_body_by_name("base_footprint"),
                _world=world,
            )

            # Create left arm
            left_gripper_thumb = Finger(
                name=PrefixedName("left_gripper_thumb", prefix=robot.name.name),
                root=world.get_body_by_name("l_gripper_l_finger_link"),
                tip=world.get_body_by_name("l_gripper_l_finger_tip_link"),
                _world=world,
            )

            left_gripper_finger = Finger(
                name=PrefixedName("left_gripper_finger", prefix=robot.name.name),
                root=world.get_body_by_name("l_gripper_r_finger_link"),
                tip=world.get_body_by_name("l_gripper_r_finger_tip_link"),
                _world=world,
            )

            left_gripper = ParallelGripper(
                name=PrefixedName("left_gripper", prefix=robot.name.name),
                root=world.get_body_by_name("l_gripper_palm_link"),
                tool_frame=world.get_body_by_name("l_gripper_tool_frame"),
                front_facing_orientation=Quaternion(0, 0, 0, 1),
                front_facing_axis=Vector3(1, 0, 0),
                thumb=left_gripper_thumb,
                finger=left_gripper_finger,
                _world=world,
            )
            left_arm = Arm(
                name=PrefixedName("left_arm", prefix=robot.name.name),
                root=world.get_body_by_name("torso_lift_link"),
                tip=world.get_body_by_name("l_wrist_roll_link"),
                manipulator=left_gripper,
                _world=world,
            )

            robot.add_arm(left_arm)

            # Create right arm
            right_gripper_thumb = Finger(
                name=PrefixedName("right_gripper_thumb", prefix=robot.name.name),
                root=world.get_body_by_name("r_gripper_l_finger_link"),
                tip=world.get_body_by_name("r_gripper_l_finger_tip_link"),
                _world=world,
            )
            right_gripper_finger = Finger(
                name=PrefixedName("right_gripper_finger", prefix=robot.name.name),
                root=world.get_body_by_name("r_gripper_r_finger_link"),
                tip=world.get_body_by_name("r_gripper_r_finger_tip_link"),
                _world=world,
            )
            right_gripper = ParallelGripper(
                name=PrefixedName("right_gripper", prefix=robot.name.name),
                root=world.get_body_by_name("r_gripper_palm_link"),
                tool_frame=world.get_body_by_name("r_gripper_tool_frame"),
                front_facing_orientation=Quaternion(0, 0, 0, 1),
                front_facing_axis=Vector3(1, 0, 0),
                thumb=right_gripper_thumb,
                finger=right_gripper_finger,
                _world=world,
            )
            right_arm = Arm(
                name=PrefixedName("right_arm", prefix=robot.name.name),
                root=world.get_body_by_name("torso_lift_link"),
                tip=world.get_body_by_name("r_wrist_roll_link"),
                manipulator=right_gripper,
                _world=world,
            )

            robot.add_arm(right_arm)

            # Create camera and neck
            camera = Camera(
                name=PrefixedName("wide_stereo_optical_frame", prefix=robot.name.name),
                root=world.get_body_by_name("wide_stereo_optical_frame"),
                forward_facing_axis=Vector3(0, 0, 1),
                field_of_view=FieldOfView(
                    horizontal_angle=0.99483, vertical_angle=0.75049
                ),
                minimal_height=1.27,
                maximal_height=1.60,
                _world=world,
            )

            neck = Neck(
                name=PrefixedName("neck", prefix=robot.name.name),
                sensors={camera},
                root=world.get_body_by_name("head_pan_link"),
                tip=world.get_body_by_name("head_tilt_link"),
                pitch_body=world.get_body_by_name("head_tilt_link"),
                yaw_body=world.get_body_by_name("head_pan_link"),
                _world=world,
            )
            robot.add_neck(neck)

            # Create torso
            torso = Torso(
                name=PrefixedName("torso", prefix=robot.name.name),
                root=world.get_body_by_name("torso_lift_link"),
                tip=world.get_body_by_name("torso_lift_link"),
                _world=world,
            )
            robot.add_torso(torso)

            world.add_semantic_annotation(robot, exists_ok=True)

            vel_limits = defaultdict(
                lambda: 1,
                {
                    world.get_connection_by_name("head_tilt_joint"): 3.5,
                    world.get_connection_by_name("r_shoulder_pan_joint"): 0.15,
                    world.get_connection_by_name("l_shoulder_pan_joint"): 0.15,
                    world.get_connection_by_name("r_shoulder_lift_joint"): 0.2,
                    world.get_connection_by_name("l_shoulder_lift_joint"): 0.2,
                },
            )
            robot.tighten_dof_velocity_limits_of_1dof_connections(new_limits=vel_limits)

        return robot



from __future__ import annotations
from abc import ABC
from dataclasses import dataclass

import numpy as np
import trimesh.boolean
from krrood.entity_query_language.predicate import (
    Predicate,
    Symbol,
)
from random_events.interval import Interval
from typing_extensions import List, TYPE_CHECKING, Iterable, Type

from ..collision_checking.trimesh_collision_detector import TrimeshCollisionDetector
from ..datastructures.prefixed_name import PrefixedName
from ..datastructures.variables import SpatialVariables
from ..spatial_computations.ik_solver import (
    MaxIterationsException,
    UnreachableException,
)
from ..spatial_computations.raytracer import RayTracer
from ..spatial_types import Vector3
from ..spatial_types.spatial_types import TransformationMatrix
from ..world import World
from ..world_description.connections import FixedConnection
from ..world_description.world_entity import Body, Region, KinematicStructureEntity

if TYPE_CHECKING:
    from ..robots.abstract_robot import (
        Camera,
    )


def stable(obj: Body) -> bool:
    """
    Checks if an object is stable in the world. Stable meaning that its position will not change after simulating
    physics in the World. This will be done by simulating the world for 10 seconds and comparing
    the previous coordinates with the coordinates after the simulation.

    :param obj: The object which should be checked
    :return: True if the given object is stable in the world False else
    """
    raise NotImplementedError("Needs multiverse")


def contact(
    body1: Body,
    body2: Body,
    threshold: float = 0.001,
) -> bool:
    """
    Checks if two objects are in contact or not.

    :param body1: The first object
    :param body2: The second object
    :param threshold: The threshold for contact detection
    :return: True if the two objects are in contact False else
    """
    tcd = TrimeshCollisionDetector(body1._world)
    result = tcd.check_collision_between_bodies(body1, body2)

    if result is None:
        return False
    return result.contact_distance < threshold


def get_visible_bodies(camera: Camera) -> List[KinematicStructureEntity]:
    """
    Get all bodies and regions that are visible from the given camera using a segmentation mask.

    :param camera: The camera for which the visible objects should be returned
    :return: A list of bodies/regions that are visible from the camera
    """
    rt = RayTracer(camera._world)
    rt.update_scene()

    # This ignores the camera orientation and sets it to identity
    cam_pose = np.eye(4, dtype=float)
    cam_pose[:3, 3] = camera.root.global_pose.to_np()[:3, 3]

    seg = rt.create_segmentation_mask(
        TransformationMatrix(cam_pose, reference_frame=camera._world.root),
        resolution=256,
    )
    indices = np.unique(seg)
    indices = indices[indices > -1]
    bodies = [camera._world.kinematic_structure[i] for i in indices]

    return bodies


def visible(camera: Camera, obj: KinematicStructureEntity) -> bool:
    """
    Checks if a body/region is visible by the given camera.
    """
    return obj in get_visible_bodies(camera)


def occluding_bodies(camera: Camera, body: Body) -> List[Body]:
    """
    Determines the bodies that occlude a given body in the scene as seen from a specified camera.

    This function uses a ray-tracing approach to check occlusion. Every body that hides anything from the target body
    is an occluding body.

    :param camera: The camera for which the occluding bodies should be returned
    :param body: The body for which the occluding bodies should be returned
    :return: A list of bodies that are occluding the given body.
    """

    # get camera pose
    camera_pose = np.eye(4, dtype=float)
    camera_pose[:3, 3] = camera.root.global_pose.to_np()[:3, 3]
    camera_pose = TransformationMatrix(camera_pose, reference_frame=camera._world.root)

    # create a world only containing the target body
    world_without_occlusion = World()
    root = Body(name=PrefixedName("root"))
    with world_without_occlusion.modify_world():
        world_without_occlusion.add_body(root)
        copied_body = Body.from_json(body.to_json())
        root_T_body = body.global_pose
        root_T_body.reference_frame = root
        root_to_copied_body = FixedConnection(
            parent=root,
            child=copied_body,
            parent_T_connection_expression=root_T_body,
        )
        world_without_occlusion.add_connection(root_to_copied_body)

    # get segmentation mask without occlusion
    ray_tracer_without_occlusion = RayTracer(world_without_occlusion)
    ray_tracer_without_occlusion.update_scene()
    segmentation_mask_without_occlusion = (
        ray_tracer_without_occlusion.create_segmentation_mask(
            camera_pose, resolution=256
        )
    )

    # get segmentation mask with occlusion
    ray_tracer_with_occlusion = RayTracer(camera._world)
    ray_tracer_with_occlusion.update_scene()
    segmentation_mask_with_occlusion = (
        ray_tracer_with_occlusion.create_segmentation_mask(camera_pose, resolution=256)
    )

    mask_without_occluders = segmentation_mask_without_occlusion[
        segmentation_mask_without_occlusion == copied_body.index
    ].nonzero()

    mask_with_occluders = segmentation_mask_with_occlusion[
        mask_without_occluders != body.index
    ]
    indices = np.unique(mask_with_occluders)
    indices = indices[indices > -1]
    bodies = [camera._world.kinematic_structure[i] for i in indices]
    return bodies


def reachable(pose: TransformationMatrix, root: Body, tip: Body) -> bool:
    """
    Checks if a manipulator can reach a given position.
    This is determined by inverse kinematics.

    :param pose: The pose to reach
    :param root: The root of the kinematic chain.
    :param tip: The threshold between the end effector and the position.
    :return: True if the end effector is closer than the threshold to the target position, False in every other case
    """
    try:
        root._world.compute_inverse_kinematics(
            root=root, tip=tip, target=pose, max_iterations=1000
        )
    except MaxIterationsException as e:
        return False
    except UnreachableException as e:
        return False
    return True


def is_supported_by(
    supported_body: Body, supporting_body: Body, max_intersection_height: float = 0.1
) -> bool:
    """
    Checks if one object is supporting another object.

    :param supported_body: Object that is supported
    :param supporting_body: Object that potentially supports the first object
    :param max_intersection_height: Maximum height of the intersection between the two objects.
    If the intersection is higher than this value, the check returns False due to unhandled clipping.
    :return: True if the second object is supported by the first object, False otherwise
    """
    if Below(supported_body, supporting_body, supported_body.global_pose)():
        return False
    bounding_box_supported_body = (
        supported_body.collision.as_bounding_box_collection_at_origin(
            TransformationMatrix(reference_frame=supported_body)
        ).event
    )
    bounding_box_supporting_body = (
        supporting_body.collision.as_bounding_box_collection_at_origin(
            TransformationMatrix(reference_frame=supported_body)
        ).event
    )

    intersection = (
        bounding_box_supported_body & bounding_box_supporting_body
    ).bounding_box()

    if intersection.is_empty():
        return False

    z_intersection: Interval = intersection[SpatialVariables.z.value]
    size = sum([si.upper - si.lower for si in z_intersection.simple_sets])
    return size < max_intersection_height


def is_body_in_region(body: Body, region: Region) -> float:
    """
    Check if the body is in the region by computing the fraction of the body's
    collision volume that lies inside the region's area volume.

    Implementation detail: both the body and region meshes are defined in their
    respective local frames; we must transform them into a common (world) frame
    using their global poses before computing the boolean intersection.

    :param body: The body for which the check should be done.
    :param region: The region to check if the body is in.
    :return: The percentage (0.0..1.0) of the body's volume that lies in the region.
    """
    # Retrieve meshes in local frames
    body_mesh_local = body.collision.combined_mesh
    region_mesh_local = region.area.combined_mesh

    # Transform copies of the meshes into the world frame
    body_mesh = body_mesh_local.copy().apply_transform(body.global_pose.to_np())
    region_mesh = region_mesh_local.copy().apply_transform(region.global_pose.to_np())
    intersection = trimesh.boolean.intersection([body_mesh, region_mesh])

    # no body volume -> zero fraction
    body_volume = body_mesh.volume
    if body_volume <= 1e-12:
        return 0.0

    return intersection.volume / body_volume


@dataclass
class SpatialRelation(Symbol, ABC):
    """
    Check if the body is spatially related to the other body if you are looking from the point of semantic annotation.
    The comparison is done using the centers of mass computed from the bodies' collision geometry.
    """

    body: Body
    """
    The body for which the check should be done.
    """

    other: Body
    """
    The other body.
     """

    point_of_semantic_annotation: TransformationMatrix
    """
    The reference spot from where to look at the bodies.
    """
    eps: float = 1e-12

    def _signed_distance_along_direction(self, index: int) -> float:
        """
        Calculate the spatial relation between self.body and self.other with respect to a given
        reference point (self.point_of_semantic_annotation) and a specified axis index. This function computes the
        signed distance along a specified direction derived from the reference point
        to compare the positions of the centers of mass of the two bodies.

        :param index: The index of the axis in the transformation matrix along which
            the spatial relation is computed.
        :return: The signed distance between the first and the second body's centers
            of mass along the given direction.
        """
        ref_np = self.point_of_semantic_annotation.to_np()
        front_world = ref_np[:3, index]
        front_norm = front_world / (np.linalg.norm(front_world) + self.eps)
        front_norm = Vector3(
            x_init=front_norm[0],
            y_init=front_norm[1],
            z_init=front_norm[2],
            reference_frame=self.point_of_semantic_annotation.reference_frame,
        )

        s_body = front_norm.dot(
            self.body.collision.center_of_mass_in_world().to_vector3()
        )
        s_other = front_norm.dot(
            self.other.collision.center_of_mass_in_world().to_vector3()
        )
        return (s_body - s_other).compile()()


@dataclass
class LeftOf(SpatialRelation):
    """
    The "left" direction is taken as the -Y axis of the given point of semantic_annotation.
    """

    def __call__(self) -> bool:
        return self._signed_distance_along_direction(1) > 0.0


@dataclass
class RightOf(SpatialRelation):
    """
    The "right" direction is taken as the +Y axis of the given point of semantic_annotation.
    """

    def __call__(self) -> bool:
        return self._signed_distance_along_direction(1) < 0.0


@dataclass
class Above(SpatialRelation):
    """
    The "above" direction is taken as the +Z axis of the given point of semantic_annotation.
    """

    def __call__(self) -> bool:
        return self._signed_distance_along_direction(2) > 0.0


@dataclass
class Below(SpatialRelation):
    """
    The "below" direction is taken as the -Z axis of the given point of semantic_annotation.
    """

    def __call__(self) -> bool:
        return self._signed_distance_along_direction(2) < 0.0


@dataclass
class Behind(SpatialRelation):
    """
    The "behind" direction is defined as the -X axis of the given point of semantic annotation.
    """

    def __call__(self) -> bool:
        return self._signed_distance_along_direction(0) < 0.0


@dataclass
class InFrontOf(SpatialRelation):
    """
    The "in front of" direction is defined as the +X axis of the given point of semantic annotation.
    """

    def __call__(self) -> bool:
        return self._signed_distance_along_direction(0) > 0.0


@dataclass
class ContainsType(Predicate):
    """
    Predicate that checks if any object in the iterable is of the given type.
    """

    iterable: Iterable
    """
    Iterable to check for objects of the given type.
    """

    obj_type: Type
    """
    Object type to check for.
    """

    def __call__(self) -> bool:
        return any(isinstance(obj, self.obj_type) for obj in self.iterable)



import os

from random_events.utils import get_full_class_name

import semantic_digital_twin.semantic_annotations.semantic_annotations
from semantic_digital_twin.utils import (
    get_semantic_digital_twin_directory_root,
    type_string_to_type,
)


def test_get_semantic_digital_twin_directory_root():
    path = os.path.abspath(__file__)
    root = get_semantic_digital_twin_directory_root(path)
    assert root == os.path.abspath(os.path.join(os.path.dirname(path), ".."))


def test_type_string_to_string():
    original_class = (
        semantic_digital_twin.semantic_annotations.semantic_annotations.Handle
    )
    original_class_name = get_full_class_name(original_class)

    converted_class = type_string_to_type(original_class_name)

    assert converted_class == original_class



from __future__ import annotations

from typing_extensions import Tuple, List, TYPE_CHECKING

import numpy as np
import trimesh
from trimesh import Scene

from ..datastructures.types import NpMatrix4x4
from ..world_description.world_entity import Body
from ..spatial_types.spatial_types import GenericSpatialType

if TYPE_CHECKING:
    from ..world import World


class RayTracer:

    world: World
    """
    The world to use for ray tracing.
    """
    _last_world_model: int
    """
    Last model version of the world to which the ray tracer was updated.
    """
    _last_world_state: int
    """
    Last state version of the world to which the ray tracer was updated.
    """
    index_to_body: dict
    """
    Maps the index of a body to the body itself.
    """
    scene_to_index: dict
    """
    Maps the index in the trimesh scene to the index of the body in the world.
    """
    scene: Scene
    """
    The trimesh scene used for ray tracing which mirrors the world.
    """

    def __init__(self, world):
        """
        Initializes the RayTracer with the given world.

        :param world: The world to use for ray tracing.
        """
        self.world = world
        self._last_world_model = -1
        self._last_world_state = -1
        self.index_to_body = {}
        self.scene_to_index = {}

        self.scene = Scene()
        self.update_scene()

    def update_scene(self):
        """
        Updates the ray tracer scene with the current state of the world.
        This method should be called whenever the world changes to ensure the ray tracer has the latest information.
        """
        if self._last_world_model != self.world._model_version:
            self.add_missing_bodies()
            self._last_world_model = self.world._model_version
        if self._last_world_state != self.world._state_version:
            self.update_transforms()
            self._last_world_state = self.world._state_version

    def add_missing_bodies(self):
        """
        Adds all bodies from the world to the ray tracer scene that are not already present.
        """
        # Bodies are added to the scene with their name as the node name plus a suffix for collision geometries.
        # We check if a body is not in the complete list of all node names in the scene graph.
        # If the body is not present, we add it to the scene.
        bodies_to_add = [
            body
            for body in self.world.bodies
            if body.name.name not in "\t".join(self.scene.graph.nodes)
        ]
        for body in bodies_to_add:
            for i, collision in enumerate(body.collision):
                self.scene.add_geometry(
                    collision.mesh,
                    node_name=body.name.name + f"_collision_{i}",
                    parent_node_name="world",
                    transform=self.world.compute_forward_kinematics_np(
                        self.world.root, body
                    )
                    @ collision.origin.to_np(),
                )
                self.scene_to_index[body.name.name + f"_collision_{i}"] = body.index
                self.index_to_body[body.index] = body

    def update_transforms(self):
        """
        Updates the transforms of all bodies in the ray tracer scene.
        This is necessary to ensure that the ray tracing uses the correct positions and orientations.
        """
        for body in self.world.bodies:
            for i, collision in enumerate(body.collision):
                transform = (
                    self.world.compute_forward_kinematics_np(self.world.root, body)
                    @ collision.origin.to_np()
                )
                self.scene.graph[body.name.name + f"_collision_{i}"] = transform

    def create_segmentation_mask(
        self, camera_pose: GenericSpatialType, resolution: int = 512
    ) -> np.ndarray:
        """
        Creates a segmentation mask for the ray tracer scene from the camera position to the target position. Each pixel
        in the mask corresponds to the index of a body in the scene or -1 if no body is hit at that pixel.

        :param camera_pose: The position of the camera.
        :param resolution: The resolution of the segmentation mask.
        :return: A segmentation mask as a numpy array.
        """
        self.update_scene()
        ray_origins, ray_directions, pixels = self.create_camera_rays(
            camera_pose, resolution=resolution
        )
        points, index_ray, index_tri = self.scene.to_mesh().ray.intersects_location(
            ray_origins, ray_directions, multiple_hits=False
        )
        bodies = self.scene.triangles_node[index_tri]

        # map the name of the scene objects to the index
        bodies = [self.scene_to_index[body] for body in bodies]

        pixel_ray = pixels[index_ray]

        # create a numpy array we can turn into an image
        # doing it with uint8 creates an `L` mode greyscale image
        a = np.zeros(self.scene.camera.resolution, dtype=np.int32) - 1

        # assign bodies to correct pixel locations
        a[pixel_ray[:, 0], pixel_ray[:, 1]] = bodies

        return a

    def create_depth_map(
        self, camera_pose: GenericSpatialType, resolution: int = 512
    ) -> np.ndarray:
        """
        Creates a depth map for the ray tracer scene from the camera position to the target position. Each pixel in the
        depth map corresponds to the distance from the camera to the closest point on the surface of the scene or -1 if
        no point is hit.

        :param camera_pose: The position of the camera.
        :param resolution: The resolution of the depth map.
        :return: A depth map as a numpy array.
        """
        self.update_scene()
        ray_origins, ray_directions, pixels = self.create_camera_rays(
            camera_pose, resolution=resolution
        )
        # Code from the example in trimesh repo: examples/raytrace.py
        points, index_ray, index_tri = self.scene.to_mesh().ray.intersects_location(
            ray_origins, ray_directions, multiple_hits=False
        )
        depth = trimesh.util.diagonal_dot(
            points - ray_origins[0], ray_directions[index_ray]
        )
        pixel_ray = pixels[index_ray]

        # create a numpy array we can turn into an image
        # doing it with uint8 creates an `L` mode greyscale image
        a = np.zeros(self.scene.camera.resolution, dtype=np.float32) - 1

        # assign depth to correct pixel locations
        a[pixel_ray[:, 0], pixel_ray[:, 1]] = depth

        return a

    def create_camera_rays(
        self, camera_pose: GenericSpatialType, resolution: int = 512, fov=90
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Creates camera rays for the ray tracer scene from the camera position to the target position. Places the camera
        at the given position and orientation view of the camera is along the x-axis.

        :param camera_pose: The position of the camera as a 4x4 transformation matrix.
        :param resolution: The resolution of the camera rays.
        :param fov: The field of view of the camera in degrees.
        :return: The origin points of the rays, the direction vectors of the rays, and the pixel coordinates.
        """
        camera_pose = camera_pose.to_np()
        self.update_scene()
        self.scene.camera.resolution = (resolution, resolution)
        # By default, the camera is looking along the -z axis, so we need to rotate it to look along the x-axis.
        rotate = trimesh.transformations.rotation_matrix(
            angle=np.radians(-90.0), direction=[0, 1, 0]
        )
        rotate_x = trimesh.transformations.rotation_matrix(
            angle=np.radians(180.0), direction=[1, 0, 0]
        )

        self.scene.camera.fov = (fov, fov)
        self.scene.camera.resolution = [resolution, resolution]
        self.scene.graph[self.scene.camera.name] = camera_pose @ rotate_x @ rotate

        return self.scene.camera_rays()

    def ray_test(
        self, origin_points: np.ndarray, target_points: np.ndarray, multiple_hits=False
    ) -> Tuple[np.ndarray, np.ndarray, List[Body]]:
        """
        Performs a ray test from the origin point to the target point in the ray tracer scene.

        :param origin_points: The starting point of the ray.
        :param target_points: The end point of the ray.
        :return: A tuple containing the points where the ray intersects and the indices of rays that hit the scene as well as the bodies that were.
        """
        origin_points = np.array(origin_points)
        target_points = np.array(target_points)
        self.update_scene()
        origin_points = origin_points.reshape((-1, 3))
        target_points = target_points.reshape((-1, 3))
        if origin_points.shape != target_points.shape:
            raise ValueError("Origin and target points must have the same shape.")

        ray_directions = target_points - origin_points
        points, index_ray, index_tri = self.scene.to_mesh().ray.intersects_location(
            origin_points, ray_directions, multiple_hits=multiple_hits
        )
        bodies = self.scene.triangles_node[index_tri]

        # map the name of the scene objects to the index
        bodies = [self.index_to_body[self.scene_to_index[body]] for body in bodies]

        return points, index_ray, bodies
